/*  Screem:  screem-tag_file.c
 *
 *  Copyright (C) 2003 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include "config.h"

#include <glib/gi18n.h>

#include "screem-tagfile.h"

#include <string.h>
#include <gtk/gtktreestore.h>
#include <gtk/gtktreemodelsort.h>
#include <gdk/gdk.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include "fileops.h"

#include "support.h"

#include "screem-file.h"

#include "screem-application.h"
#include "screem-icon-cache.h"

extern ScreemApplication *app;

static gint screem_tag_file_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, 
					  GtkTreeIter *b,
				    	  gpointer data );
static void screem_tag_file_icon_change( ScreemIconCache *cache, 
					gpointer data );
static void screem_tag_file_free_node_info( NodeInfo *info );
static void screem_tag_file_clear( ScreemTagFile *tagfile );
static gboolean screem_tag_file_clear_foreach( GtkTreeModel *model,
		GtkTreePath *path, GtkTreeIter *iter, gpointer data );


enum {
	PROP_0,
	PROP_ACTIVE
};

struct ScreemTagFilePrivate {
	ScreemTagFile *tfile;
	ScreemFile *file;

	GStringChunk *chunk;
	
	gchar *filename;

	gchar *name;
	gchar *description;

	gchar *type;

	ScreemIconCache *cache;
	
	/* the unsorted browser model */
	GtkTreeModel *model;
	
	gboolean active;

	gboolean loaded;

	/* current auto complete prefix */
	const gchar *prefix;
	guint prefix_len;

	ScreemTagFileFormat format;

	/* current parent path */
	GtkTreePath *path;

	/* tmp pixbuf for folder icon to avoid lots of
	 * cache lookups */
	GdkPixbuf *pixbuf;


	xmlXPathCompExprPtr text_exp;

	xmlXPathCompExprPtr desc_exp_l;
	xmlXPathCompExprPtr ins_exp_l;
	xmlXPathCompExprPtr prop_exp_l;
	xmlXPathCompExprPtr group_exp_l;
	xmlXPathCompExprPtr entry_exp_l;

	
	xmlXPathCompExprPtr desc_exp;
	xmlXPathCompExprPtr ins_exp;
	xmlXPathCompExprPtr prop_exp;
	xmlXPathCompExprPtr group_exp;
	xmlXPathCompExprPtr entry_exp;
};

G_DEFINE_TYPE( ScreemTagFile, screem_tag_file, G_TYPE_OBJECT )

static void screem_tag_file_finalize( GObject *object );
static void screem_tag_file_set_prop( GObject *object, guint prop_id,
				const GValue *value, GParamSpec *spec );
static void screem_tag_file_get_prop( GObject *object, guint prop_id,
				GValue *value, GParamSpec *spec );


static void screem_tag_file_class_init( ScreemTagFileClass *klass )
{
	GObjectClass *obj_class;
	GParamSpec *spec;

	obj_class = G_OBJECT_CLASS( klass );
	obj_class->finalize = screem_tag_file_finalize;
	obj_class->get_property = screem_tag_file_get_prop;
	obj_class->set_property = screem_tag_file_set_prop;

	spec = g_param_spec_boolean( "active", "active",
				_( "Is the Tag File active or not" ),
				TRUE, 
				G_PARAM_READABLE | G_PARAM_WRITABLE );
	g_object_class_install_property( obj_class, PROP_ACTIVE,
					spec );
}

static void screem_tag_file_init( ScreemTagFile *tag_file )
{
	ScreemTagFilePrivate *priv;
	GtkTreeStore *store;
	priv = tag_file->priv = g_new0( ScreemTagFilePrivate, 1 );
	priv->tfile = tag_file;

	/* used for attribute names function names etc.
	 * alloc a reasonable size */
	priv->chunk = g_string_chunk_new( 20 );
	
	priv->cache = screem_application_get_icon_cache( app );

	store = gtk_tree_store_new( SCREEM_TAG_FILE_COLS,
			GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_POINTER );
	priv->model = GTK_TREE_MODEL( store );
	
	g_object_set_data( G_OBJECT( priv->model ), "tagfile", tag_file );
	
	gtk_tree_sortable_set_sort_func( GTK_TREE_SORTABLE( store ), 
			SCREEM_TAG_FILE_NAME,
			(GtkTreeIterCompareFunc)screem_tag_file_compare_func, 
			tag_file, NULL );
	
	g_signal_connect( G_OBJECT( priv->cache ), "changed",
		G_CALLBACK( screem_tag_file_icon_change ), tag_file );


	priv->text_exp = xmlXPathCompile( (const xmlChar*)"text()" );

	priv->desc_exp_l = xmlXPathCompile( (const xmlChar*)"description[(@xml:lang)]" );
	priv->ins_exp_l = xmlXPathCompile( (const xmlChar*)"insert[(@xml:lang)]" );
	priv->prop_exp_l = xmlXPathCompile( (const xmlChar*)"(property|param)[(@xml:lang)]" );
	priv->group_exp_l = xmlXPathCompile( (const xmlChar*)"group[(@xml:lang)]" );
	priv->entry_exp_l = xmlXPathCompile( (const xmlChar*)"(entry|function)[(@xml:lang)]" );

	priv->desc_exp = xmlXPathCompile( (const xmlChar*)"description[not(@xml:lang)]" );
	priv->ins_exp = xmlXPathCompile( (const xmlChar*)"insert[not(@xml:lang)]" );
	priv->prop_exp = xmlXPathCompile( (const xmlChar*)"(property|param)[not(@xml:lang)]" );
	priv->group_exp = xmlXPathCompile( (const xmlChar*)"group[not(@xml:lang)]" );
	priv->entry_exp = xmlXPathCompile( (const xmlChar*)"(entry|function)[not(@xml:lang)]" );

}

static void screem_tag_file_finalize( GObject *object )
{
	ScreemTagFile *tag_file;
	ScreemTagFilePrivate *priv;
		
	g_return_if_fail( object != NULL );
	g_return_if_fail( SCREEM_IS_TAGFILE( object ) );

	tag_file = SCREEM_TAGFILE( object );

	priv = tag_file->priv;

	screem_tag_file_clear( tag_file );
	g_string_chunk_free( priv->chunk );

	g_free( priv->filename );
	g_free( priv->name );
	g_free( priv->description );
	g_free( priv->type );
	
	g_object_unref( priv->cache );

	xmlXPathFreeCompExpr( priv->text_exp );

	xmlXPathFreeCompExpr( priv->desc_exp_l );
	xmlXPathFreeCompExpr( priv->desc_exp );
	xmlXPathFreeCompExpr( priv->ins_exp_l );
	xmlXPathFreeCompExpr( priv->ins_exp );
	xmlXPathFreeCompExpr( priv->prop_exp_l );
	xmlXPathFreeCompExpr( priv->prop_exp );
	xmlXPathFreeCompExpr( priv->group_exp_l );
	xmlXPathFreeCompExpr( priv->group_exp );
	xmlXPathFreeCompExpr( priv->entry_exp_l );
	xmlXPathFreeCompExpr( priv->entry_exp );

	
	g_free( priv );
	
	G_OBJECT_CLASS( screem_tag_file_parent_class )->finalize( object );
}

static void screem_tag_file_set_prop( GObject *object, guint prop_id,
		const GValue *value, GParamSpec *spec )
{
	ScreemTagFile *tag_file;
	ScreemTagFilePrivate *priv;
	
	tag_file = SCREEM_TAGFILE( object );
	priv = tag_file->priv;

	switch( prop_id ) {
		case PROP_ACTIVE:
			priv->active = g_value_get_boolean( value );
			if( priv->active && ! priv->loaded ) {
				screem_tag_file_load( tag_file,
						priv->format );
			}
			break;
		default:
			break;
	}
}

static void screem_tag_file_get_prop( GObject *object, guint prop_id,
		GValue *value, GParamSpec *spec )
{
	ScreemTagFile *tag_file;
	ScreemTagFilePrivate *priv;
	
	tag_file = SCREEM_TAGFILE( object );
	priv = tag_file->priv;

	switch( prop_id ) {
		case PROP_ACTIVE:
			g_value_set_boolean( value, priv->active );
			break;
		default:
			break;
	}
}

/* static stuff */
static gint screem_tag_file_compare_func( GtkTreeModel *model, 
				    	  GtkTreeIter *a, 
					  GtkTreeIter *b,
				    	  gpointer data )
{
        NodeInfo *ainfo;
        NodeInfo *binfo;
	int ret = 0;

	gtk_tree_model_get( model, a, 
			SCREEM_TAG_FILE_DATA, &ainfo, -1 );
	gtk_tree_model_get( model, b, 
			SCREEM_TAG_FILE_DATA, &binfo, -1 );

	if( ainfo && ! binfo ) {
		ret = 1;
	} else if( binfo && !ainfo ) {
		ret = -1;
	} else if( ! ainfo && ! binfo ) {
		ret = 0;
	} else {
		ret = strcmp( ainfo->name, binfo->name );
	}
	
	return ret;
}

static void screem_tag_file_foreach( GtkTreeModel *model, 
		GtkTreePath *path, GtkTreeIter *it,
		ScreemTagFile *tfile )
{
	NodeInfo *info;
	ScreemTagFilePrivate *priv;
	ScreemIconCache *cache;
	GdkPixbuf *pixbuf;
	
	priv = tfile->priv;
	cache = priv->cache;
	
	gtk_tree_model_get( model, it, 
			SCREEM_TAG_FILE_DATA, &info, -1 );
	if( info ) {
		pixbuf = NULL;
		if( info->file ) {
			pixbuf = screem_icon_cache_get_pixbuf( cache,
					info->file, NULL, 
					SCREEM_ICON_CACHE_DEFAULT_SIZE );

		} else if( gtk_tree_model_iter_has_child( model, it ) ) {
			pixbuf = priv->pixbuf;
		}
		
		gtk_tree_store_set( GTK_TREE_STORE( model ), it,
				SCREEM_TAG_FILE_ICON, pixbuf, -1 );
		if( pixbuf && info->file ) {
			g_object_unref( pixbuf );
		}
	}

}

static void screem_tag_file_icon_change( ScreemIconCache *cache, 
					gpointer data )
{
	ScreemTagFile *tfile;
	ScreemTagFilePrivate *priv;
	GtkTreeModel *model;
	
	tfile = SCREEM_TAGFILE( data );
	priv = tfile->priv;
	model = tfile->priv->model;

	priv->pixbuf = screem_icon_cache_get_pixbuf( cache,
			"/", "x-directory/normal",
			SCREEM_ICON_CACHE_DEFAULT_SIZE );

	gtk_tree_model_foreach( model, 
			(GtkTreeModelForeachFunc)screem_tag_file_foreach,
			tfile );

	if( priv->pixbuf ) {
		g_object_unref( priv->pixbuf );
		priv->pixbuf = NULL;
	}
}

static void screem_tag_file_free_node_info( NodeInfo *info )
{
	GSList *list;
	NodeAttribute *attr;

	/* info->name is in the string chunk */
	g_free( info->open );
	g_free( info->close );
	g_free( info->file );
	g_free( info->text );
	g_free( info->description );

	for( list = info->attributes; list; list = list->next ) {
		attr = (NodeAttribute*)list->data;

		/* attr->name is in the string chunk */
		
		g_free( attr->type );
		g_free( attr->defvalue );
		g_free( attr->description );

		/* attr is a single block of memory, so don't free each
		 * item in the list unless this is the start of that
		 * block */
		if( ! list->next ) {
			g_free( attr );
		}
	}
	if( info->attributes ) {
		g_slist_free( info->attributes );
	}
	
	g_free( info );
}

static void screem_tag_file_clear( ScreemTagFile *tagfile )
{
	ScreemTagFilePrivate *priv;
	
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );

	priv = tagfile->priv;
	
	gtk_tree_model_foreach( priv->model,
			screem_tag_file_clear_foreach,
			NULL );

	gtk_tree_store_clear( GTK_TREE_STORE( priv->model ) );

	g_string_chunk_free( priv->chunk );
	priv->chunk = g_string_chunk_new( 20 );
}

static gboolean screem_tag_file_clear_foreach( GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data )
{
	NodeInfo *info;
	
	gtk_tree_model_get( model, iter, 
			SCREEM_TAG_FILE_DATA, &info, -1 );
	if( info ) {
		screem_tag_file_free_node_info( info );
		gtk_tree_store_set( GTK_TREE_STORE( model ), iter,
				SCREEM_TAG_FILE_DATA, NULL, -1 );
	} else {
		g_warning( "NULL DATA COL in Tag Tree\n" );
	}

	return FALSE;
}

static gboolean screem_tag_file_autocomplete_fill( GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data )
{
	ScreemTagFilePrivate *priv;
	GSList **ret;
	ScreemTagFile *tfile;
	NodeInfo *info;
	const gchar *str;
	const gchar *prefix;
	guint len;

	tfile = SCREEM_TAGFILE( g_object_get_data( G_OBJECT( model ),
						"tagfile" ) );
	priv = tfile->priv;
	prefix = priv->prefix;
	len = priv->prefix_len;
	ret = (GSList**)data;

	gtk_tree_model_get( model, iter,
			SCREEM_TAG_FILE_DATA, &info, -1 );

	str = info->open;
	if( ! str ) {
		str = info->close;
	}
	if( ! str ) {
		str = info->text;
	}
	if( str && 
	   ! strncmp( str, prefix, len ) ) {
		*ret = g_slist_prepend( *ret, g_strdup( str ) );
	}

	return FALSE;
}

static gboolean screem_tag_file_autocomplete_tip_fill( GtkTreeModel *model,
						GtkTreePath *path,
						GtkTreeIter *iter,
						gpointer data )
{
	ScreemTagFilePrivate *priv;
	GSList **ret;
	ScreemTagFile *tfile;
	NodeInfo *info;
	const gchar *str;
	const gchar *prefix;
	guint len;
	
	tfile = SCREEM_TAGFILE( g_object_get_data( G_OBJECT( model ),
						"tagfile" ) );
	priv = tfile->priv;
	prefix = priv->prefix;
	len = priv->prefix_len;
	ret = (GSList**)data;

	g_return_val_if_fail( ret != NULL, FALSE );
	
	gtk_tree_model_get( model, iter, 
			SCREEM_TAG_FILE_DATA, &info, -1 );

	str = info->tip;
	if( ! str ) {
		str = info->open;
	}
	if( ! str ) {
		str = info->close;
	}
	if( ! str ) {
		str = info->text;
	}
	if( str && 
	   ! strncmp( str, prefix, len ) ) {
		*ret = g_slist_prepend( *ret, (gchar*)str );
	}

	return FALSE;
}

static gboolean lookup_lang( const gchar *find, const gchar * const *langs )
{
	gboolean ret;
	const gchar *lang;
	
	ret = FALSE;
	if( langs ) {
		while( ! ret && ( lang = *langs ) ) {
			ret = ! strcmp( find, lang );
		}
	}	

	return ret;
}

static gboolean find_node( xmlXPathCompExprPtr expr, 
		xmlXPathCompExprPtr fallback, 
		gboolean lang,
		xmlXPathContextPtr ctx,
		xmlNodePtr *node, xmlXPathObjectPtr *o )
{
	xmlXPathObjectPtr obj;
	xmlNodeSetPtr nodes;
	xmlNodePtr n;
	const gchar *const *langs;
	gint i;
	gchar *xmllang;
	gboolean found;

	langs = g_get_language_names();

	if( ! node ) {
		node = &n;
	}
	if( o ) {
		*o = NULL;
	}
	
	/* first we want a <ref> matching the locale */
	*node = NULL;
	obj = NULL;
	found = FALSE;
	/* lang will be TRUE if we are inside a parent element
	 * that has an xml:lang set */
	if( ! lang ) {
		obj = xmlXPathCompiledEval( expr, ctx );
	}
	if( obj ) {
		/* got some results */
		nodes = obj->nodesetval;
		for( i = 0; ( ( *node = xmlXPathNodeSetItem( nodes, i ) ) ); ++ i ) {
			xmllang = (gchar*)xmlGetNsProp( *node, 
					(const xmlChar*)"lang",
					(const xmlChar*)XML_XML_NAMESPACE );
			found = lookup_lang( (const gchar*)xmllang, langs );
			g_free( xmllang );
			if( found ) {
				if( o ) {
					*o = obj;
				}
				break;
			}
		}
		if( ( ! *node ) || ! o ) {
			xmlXPathFreeObject( obj );
		}
	}
	if( ! *node ) {
		/* failed to find one, fall back */
		obj = xmlXPathCompiledEval( fallback, ctx );
		if( obj ) {
			nodes = obj->nodesetval;
			*node = xmlXPathNodeSetItem( nodes, 0 );
			if( o ) {
				*o = obj;
			} else {
				xmlXPathFreeObject( obj );
			}
		}
	}

	return found;
}

static xmlNodePtr find_text( xmlXPathContextPtr ctx,
		ScreemTagFile *tfile )
{
	xmlXPathObjectPtr obj;
	xmlNodePtr snode;

	obj = xmlXPathCompiledEval( tfile->priv->text_exp, ctx );
	snode = NULL;
	if( obj ) {
		snode = xmlXPathNodeSetItem( obj->nodesetval, 0 );
		xmlXPathFreeObject( obj );
	}

	return snode;
}

static void process_param( xmlXPathContextPtr ctx, xmlNodePtr node, 
		gboolean lang, gpointer userdata )
{
	ScreemTagFile *tfile;
	ScreemTagFilePrivate *priv;
	xmlNodePtr snode;

	gchar *name;
	gchar *desc;
	gchar *required;
	gchar *vallist;
	gchar *defval;
	gchar *type;

	GtkTreeIter it;
	NodeInfo *info;
	NodeAttribute *attr;

	tfile = SCREEM_TAGFILE( userdata );
	priv = tfile->priv;
	
	/* process attrs */
	name = (gchar*)xmlGetProp( node, (const xmlChar*)"name" );
	required = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"required" );
	vallist = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"vallist" );
	defval = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"default" );
	type = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"type" );
	desc = NULL;

	/* process description */
	ctx->node = node;

	snode = find_text( ctx, tfile );
	if( snode ) {
		desc = (gchar*)snode->content;
	}

	gtk_tree_model_get_iter( priv->model, &it, priv->path );
	gtk_tree_model_get( priv->model, &it,
			SCREEM_TAG_FILE_DATA, &info, -1 );
	
	attr = info->attributes->data;
	if( name ) {
		attr->name =  g_string_chunk_insert_const( priv->chunk,
			gettext( name ) );
		xmlFree( name );
	}
	if( required ) {
		attr->required = ( *required == '1' );
		xmlFree( required );
	}
	if( vallist ) {
		attr->vallist = ( *vallist == '1' );
		xmlFree( vallist );
	}
	if( defval ) {
		attr->defvalue = g_strdup( defval );
		xmlFree( defval );
	}
	if( type ) {
		attr->type = g_strdup( type );
		xmlFree( type );
	}
	if( desc ) {
		attr->description = g_strdup( gettext( desc ) );
		attr->description = g_strstrip( attr->description );
	}
}

static void process_entry( xmlXPathContextPtr ctx, xmlNodePtr node, 
		gboolean lang, gpointer userdata )
{
	ScreemTagFile *tfile;
	ScreemTagFilePrivate *priv;
	xmlXPathObjectPtr obj;
	xmlNodeSetPtr set;
	xmlNodePtr snode;
	gboolean slang;
	gint i;

	gchar *name;
	gchar *desc;

	gchar *open;
	gchar *close;
	gchar *href;
	gchar *data;
	gchar *pathname;
	
	GtkTreeIter it;
	GtkTreeIter pit;
	GdkPixbuf *pixbuf;
	NodeInfo *info;
	GtkTreePath *path;

	gboolean isfunc;
	
	tfile = SCREEM_TAGFILE( userdata );
	priv = tfile->priv;
	
	isfunc = ( ! strcmp( "function", (gchar*)node->name ) );
	
	/* process attrs */
	name = (gchar*)xmlGetProp( node, (const xmlChar*)"name" );
	desc = NULL;

	/* process description */
	ctx->node = node;
	slang = find_node( priv->desc_exp_l, priv->desc_exp,
			lang, ctx, &snode, NULL );
	if( snode ) {
		/* get text */
		ctx->node = snode;
		snode = find_text( ctx, tfile );
		if( snode ) {
			desc = (gchar*)snode->content;
		}
	}

	pixbuf = NULL;
	
	info = g_new0( NodeInfo, 1 );
	info->name = g_string_chunk_insert_const( priv->chunk,
			gettext( name ) );
	g_free( name );
	info->isfunc = isfunc;
	if( desc ) {
		info->description = g_strdup( gettext( desc ) );
	}
	
	gtk_tree_model_get_iter( priv->model, &pit, priv->path );
	gtk_tree_store_append( GTK_TREE_STORE( priv->model ), 
			&it, &pit );
	gtk_tree_store_set( GTK_TREE_STORE( priv->model ), &it,
			SCREEM_TAG_FILE_NAME, info->name,
			SCREEM_TAG_FILE_ICON, pixbuf,
			SCREEM_TAG_FILE_DATA, info,
			-1 );
	path = gtk_tree_model_get_path( priv->model, &it );


	/* process insert */
	ctx->node = node;
	slang = find_node( priv->ins_exp_l, priv->ins_exp, 
			lang, ctx, &snode, NULL );
	if( snode ) {
		/* process attrs */
		open = (gchar*)xmlGetProp( snode, (const xmlChar*)"open" );
		close = (gchar*)xmlGetProp( snode, 
				(const xmlChar*)"close" );
		href = (gchar*)xmlGetProp( snode, 
				(const xmlChar*)"href" );

		/* get text */
		ctx->node = snode;
		snode = find_text( ctx, tfile );
		data = NULL;
		if( snode ) {
			data = g_strdup( (gchar*)snode->content );
		}

		if( href ) {
			pathname = relative_to_full( href,
					RESOURCE_PATH );
			info->file = pathname;
			pixbuf = screem_icon_cache_get_pixbuf( priv->cache,
					pathname, NULL,
					SCREEM_ICON_CACHE_DEFAULT_SIZE );
		} else {
			if( open ) {
				/* open text */
				info->open = g_strdup( open );
			} 
			if( close ) {
				/* close text */
				info->close = g_strdup( close );
			}
		}
		if( isfunc ) {
			info->text = g_strconcat( info->name, "()",
					NULL );
			if( data ) {
				info->tip = g_strstrip( data );
			}
		} else if( data ) {
			info->text = g_strstrip( data );
		}
		g_free( open );
		g_free( close );
		g_free( href );
	}
	
	/* process property / param */
	ctx->node = node;
	slang = find_node( priv->prop_exp_l, priv->prop_exp, 
			lang, ctx, NULL, &obj );
	if( obj ) {
		NodeAttribute *attrs;
		
		set = obj->nodesetval;

		/* alloc a single chunk of memory for all nodes */
		attrs = g_new0( NodeAttribute, set->nodeNr );
		
		for( i = 0; i < set->nodeNr; ++ i ) {
			snode = set->nodeTab[ i ];

			info->attributes = g_slist_prepend( info->attributes, &attrs[ i ] );
			/* process <property> / <param> */
			priv->path = path;
			process_param( ctx, snode, lang, userdata );
		}
		xmlXPathFreeObject( obj );
	}

	gtk_tree_store_set( GTK_TREE_STORE( priv->model ), &it,
			SCREEM_TAG_FILE_ICON, pixbuf, -1 );

	if( pixbuf ) {
		g_object_unref( pixbuf );
	}
	gtk_tree_path_free( path );
}

static void process_group( xmlXPathContextPtr ctx, xmlNodePtr node, 
	gboolean lang, gpointer userdata )
{
	ScreemTagFile *tfile;
	ScreemTagFilePrivate *priv;
	xmlXPathObjectPtr obj;
	xmlNodeSetPtr set;
	xmlNodePtr snode;
	gboolean slang;
	gint i;

	gchar *name;
	gchar *desc;

	GtkTreeIter it;
	GtkTreeIter pit;
	GdkPixbuf *pixbuf;
	NodeInfo *info;
	GtkTreePath *path;
	
	tfile = SCREEM_TAGFILE( userdata );
	priv = tfile->priv;
	
	/* process attrs */
	name = (gchar*)xmlGetProp( node, (const xmlChar*)"name" );
	desc = (gchar*)xmlGetProp( node, (const xmlChar*)"description" );
	
	/* process description */
	if( ! desc ) {
		ctx->node = node;
		slang = find_node( priv->desc_exp_l, priv->desc_exp,
				lang, ctx, &snode, NULL );
		if( snode ) {
			/* get text */
			ctx->node = snode;
			snode = find_text( ctx, tfile );
			if( snode ) {
				desc = g_strdup( (gchar*)snode->content );
			}
		}
	}

	pixbuf = priv->pixbuf;
	
	info = g_new0( NodeInfo, 1 );
	info->name =  g_string_chunk_insert_const( priv->chunk,
			gettext( name ) );
	g_free( name );
	if( desc ) {
		info->description = desc;
	}
	
	gtk_tree_model_get_iter( priv->model, &pit, priv->path );
	gtk_tree_store_append( GTK_TREE_STORE( priv->model ), 
			&it, &pit );
	gtk_tree_store_set( GTK_TREE_STORE( priv->model ), &it,
			SCREEM_TAG_FILE_NAME, info->name,
			SCREEM_TAG_FILE_ICON, pixbuf,
			SCREEM_TAG_FILE_DATA, info,
			-1 );
	path = gtk_tree_model_get_path( priv->model, &it );

	/* get subgroups */
	ctx->node = node;
	slang = find_node( priv->group_exp_l, priv->group_exp,
			lang, ctx, NULL, &obj );
	if( obj && ( obj->nodesetval->nodeNr != 0 ) ) {
		set = obj->nodesetval;
		for( i = 0; i < set->nodeNr; ++ i ) {
			snode = set->nodeTab[ i ];

			/* process <group> */
			priv->path = path;
			process_group( ctx, snode, slang, userdata );
		}
		xmlXPathFreeObject( obj );
	} else {
		if( obj ) {
			xmlXPathFreeObject( obj );
		}
		/* get entries */
		ctx->node = node;
		slang = find_node( priv->entry_exp_l, priv->entry_exp, 
				lang, ctx, NULL, &obj );
		if( obj ) {
			set = obj->nodesetval;
			for( i = 0; i < set->nodeNr; ++ i ) {
				snode = set->nodeTab[ i ];

				/* process <entry> */
				priv->path = path;
				process_entry( ctx, snode, slang,
						userdata );
			}
			xmlXPathFreeObject( obj );
		}
	}

	gtk_tree_path_free( path );
}

static void process_ref( xmlXPathContextPtr ctx, xmlNodePtr node,
		gboolean lang, gpointer userdata )
{
	ScreemTagFile *tfile;
	ScreemTagFilePrivate *priv;
	xmlXPathObjectPtr obj;
	xmlNodeSetPtr set;
	xmlNodePtr gnode;
	gint i;

	gchar *name;
	gchar *desc;
	gchar *type;
	
	GtkTreeIter it;
	GdkPixbuf *pixbuf;
	NodeInfo *info;
	GtkTreePath *path;
	gboolean slang;
	
	tfile = SCREEM_TAGFILE( userdata );
	priv = tfile->priv;
	
	/* process attrs */
	name = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"name" );
	desc = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"description" );
	type = (gchar*)xmlGetProp( node, 
			(const xmlChar*)"type" );

	g_free( priv->name );
	g_free( priv->description );
	g_free( priv->type );
	priv->name = g_strdup( gettext( name ) );
	g_free( name );
	if( desc ) {
		priv->description = desc;
	}
	if( type ) {
		priv->type = type;
	}

	pixbuf = priv->pixbuf = screem_icon_cache_get_pixbuf( priv->cache,
			"/", "x-directory/normal", 
			SCREEM_ICON_CACHE_DEFAULT_SIZE );
	
	info = g_new0( NodeInfo, 1 );
	info->name =  g_string_chunk_insert_const( priv->chunk,
			priv->name );
	info->tfile = tfile;
	if( desc ) {
		info->description = g_strdup( desc );
	}

	gtk_tree_store_append( GTK_TREE_STORE( priv->model ), &it,
			NULL );
	gtk_tree_store_set( GTK_TREE_STORE( priv->model ), &it,
			SCREEM_TAG_FILE_NAME, info->name,
			SCREEM_TAG_FILE_ICON, pixbuf,
			SCREEM_TAG_FILE_DATA, info,
			-1 );
	
	path = gtk_tree_model_get_path( priv->model, &it );
	
	ctx->node = node;
	
	/* get groups */
	lang = find_node( priv->group_exp_l, priv->group_exp,
			lang, ctx, NULL, &obj );
	if( obj  && ( obj->nodesetval->nodeNr != 0 ) ) {
		set = obj->nodesetval;
		for( i = 0; i < set->nodeNr; ++ i ) {
			gnode = set->nodeTab[ i ];
			
			/* process <group> */
			priv->path = path;
			process_group( ctx, gnode, lang, userdata );
		}
		xmlXPathFreeObject( obj );
	} else {
		/* hmm no groups, cheat and treat <ref> as a group */
		if( obj ) {
			xmlXPathFreeObject( obj );
		}
		
		/* get entries */
		ctx->node = node;
		slang = find_node( priv->entry_exp_l, priv->entry_exp, 
				lang, ctx, NULL, &obj );
		if( obj ) {
			set = obj->nodesetval;
			for( i = 0; i < set->nodeNr; ++ i ) {
				gnode = set->nodeTab[ i ];

				/* process <entry> */
				priv->path = path;
				process_entry( ctx, gnode, slang,
						userdata );
			}
			xmlXPathFreeObject( obj );
		}
	}

	if( pixbuf ) {
		g_object_unref( G_OBJECT( pixbuf ) );
		priv->pixbuf = NULL;
	}

	gtk_tree_path_free( path );
	priv->path = NULL;
}

static gboolean parse_from_doc( xmlDocPtr doc, gpointer userdata )
{
	xmlXPathContextPtr ctx;
	xmlNodePtr node;
	gboolean lang;

	if( doc ) {
		xmlXPathCompExprPtr expr;
		xmlXPathCompExprPtr fallback;

		ctx = xmlXPathNewContext( doc );
		
		/* bluefish doesn't have refs as it doesn't support
		 * inline translations so we need to look for /ref
		 * as well */
		expr = xmlXPathCtxtCompile( ctx, (const xmlChar*) 
				"(/refs/ref|/ref)[@xml:lang]" );
		fallback = xmlXPathCtxtCompile( ctx, (const xmlChar*)
				"(/refs/ref|/ref)[not(@xml:lang)]" );
		
		lang = find_node( expr, fallback,
				FALSE, ctx, &node, NULL );

		xmlXPathFreeCompExpr( expr );
		xmlXPathFreeCompExpr( fallback );
		
		if( node ) {
			/* process <ref> */
			process_ref( ctx, node, lang, userdata );
		} 
	
		xmlXPathFreeContext( ctx );
		xmlFreeDoc( doc );
		if( ! node ) {
			doc = NULL;
		}
	}

	return ( doc != NULL );
}

static gboolean parse_from_file( const gchar *uri, gpointer userdata )
{
	xmlDocPtr doc;

	doc = xmlParseFile( uri );
	
	return parse_from_doc( doc, userdata );
}

static gboolean parse_from_string( const gchar *buffer, guint len,
		gpointer userdata )
{
	xmlDocPtr doc;
	
	doc = xmlParseMemory( buffer, len );
	
	return parse_from_doc( doc, userdata );
}

static void complete( ScreemFile *file, const gchar *uri, 
		gboolean success, ScreemTagFile *tagfile )
{
	const gchar *data;
	guint64 size;

	if( success ) {
		data = screem_file_get_data( file );
		if( data ) {
			size = screem_file_get_size( file );
			screem_tag_file_set_from_string( tagfile, 
					data, size );
		}
	} else {
		GError *error;

		error = screem_file_get_error( file );
		g_print( "Error %s = %s\n", uri, error->message );
		g_error_free( error );
	}
	g_object_set( G_OBJECT( tagfile ), "active", success, NULL );
//	g_object_unref( file );
}

/* public stuff */

ScreemTagFile *screem_tag_file_new( void )
{
	ScreemTagFile *tag_file;

	tag_file = g_object_new( SCREEM_TYPE_TAGFILE, NULL );

	return tag_file;
}

void screem_tag_file_load( ScreemTagFile *tagfile,
		ScreemTagFileFormat format )
{
	ScreemTagFilePrivate *priv;
	gchar *type;
	
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );

	priv = tagfile->priv;
	
	if( format == SCREEM_TAG_UNKNOWN_FORMAT ) {
		type = screem_get_mime_type( priv->filename,
				FALSE ); 
		if( ! strcmp( "application/x-screem-tag-tree",
					type ) ) {
			format = SCREEM_TAG_TREE_FORMAT;
		} else if( ! strcmp( "application/x-weml+xml",
					type ) ) {
			format = SCREEM_TAG_WEML_FORMAT;
		} else if( ! strcmp( "text/xml", type ) ) {
			/* treat text/xml as screem tag trees,
			 * this is as bluefish uses .xml and
			 * we support those files */
			format = SCREEM_TAG_TREE_FORMAT;
		} else {
			g_free( type );
			return;
		}
		g_free( type );
	}
	priv->format = format;

	priv->file = screem_file_new_with_uri( priv->filename );
	g_signal_connect( G_OBJECT( priv->file ), "complete",
			G_CALLBACK( complete ), tagfile );
	screem_file_load( priv->file );
	
	/* unref so it will be destroyed on completion */
	g_object_unref( priv->file );
}

void screem_tag_file_load_with_uri( ScreemTagFile *tagfile,
				const gchar *uri,
				ScreemTagFileFormat format )
{
	screem_tag_file_set_uri( tagfile, uri );

	screem_tag_file_load( tagfile, format );
}

gboolean screem_tag_file_set_from_string( ScreemTagFile *tagfile,
					const gchar *str,
					guint length )
{
	ScreemTagFilePrivate *priv;
	gboolean ret;

	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), FALSE );
	g_return_val_if_fail( str != NULL, FALSE );
	
	priv = tagfile->priv;
	
	if( length == 0 ) {
		length = strlen( str );
	}
	
	ret = FALSE;

	screem_tag_file_clear( tagfile );
	
	switch( priv->format ) {
		case SCREEM_TAG_WEML_FORMAT:

			break;
		default:
			ret = parse_from_string( str, length, tagfile );
			break;
	}
	
	priv->loaded = ret;
	
	return ret;	
}

GtkTreeModel *screem_tag_file_get_model( ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->model;
}

const gchar *screem_tag_file_get_name( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->name;
}

const gchar *screem_tag_file_get_desc( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->description;
}

const gchar *screem_tag_file_get_uri( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->filename;
}

const gchar *screem_tag_file_for_type( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );

	return tagfile->priv->type;
}

ScreemTagFileFormat screem_tag_file_get_format( const ScreemTagFile *tagfile )
{
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), 0 );

	return tagfile->priv->format;
}

void screem_tag_file_set_name( ScreemTagFile *tagfile, 
				const gchar *name )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );
	g_return_if_fail( name != NULL );

	g_free( tagfile->priv->name );
	tagfile->priv->name = g_strdup( name );
}

void screem_tag_file_set_desc( ScreemTagFile *tagfile,
				const gchar *desc )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );
	g_return_if_fail( desc != NULL  );

	g_free( tagfile->priv->description );
	tagfile->priv->description = g_strdup( desc );
}

void screem_tag_file_set_uri( ScreemTagFile *tagfile,
				const gchar *uri )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );
	g_return_if_fail( uri != NULL );

	g_free( tagfile->priv->filename );
	tagfile->priv->filename = g_strdup( uri );
}

void screem_tag_file_set_format( ScreemTagFile *tagfile,
		ScreemTagFileFormat format )
{
	g_return_if_fail( SCREEM_IS_TAGFILE( tagfile ) );

	tagfile->priv->format = format;
}

GSList *screem_tag_file_autocomplete( ScreemTagFile *tagfile,
		const gchar *prefix, gboolean fortip )
{
	ScreemTagFilePrivate *priv;
	GSList *ret;
	GtkTreeModelForeachFunc func;
	
	g_return_val_if_fail( SCREEM_IS_TAGFILE( tagfile ), NULL );
	g_return_val_if_fail( prefix != NULL, NULL );

	priv = tagfile->priv;
	
	ret = NULL;
	priv->prefix = prefix;
	priv->prefix_len = strlen( prefix );
	if( ! fortip ) {
		func = screem_tag_file_autocomplete_fill;
	} else {
		func = screem_tag_file_autocomplete_tip_fill;
	}
	gtk_tree_model_foreach( priv->model, func, &ret );

	return ret;			
}

